<?php

// interpretes sets of Gerber layers
// requires GM1, GM2 and GM3
class MKS_GerberInterpreter
{
	// arrays with an element for each layer
	protected $layerName = array(		// name / file extension, used for reference and debugging
		0 => 'GM1', 					// Outline
		1 => 'GM2', 					// Outline Single
		2 => 'GM3');					// V-Cut
	protected $nodeSets = array();		// number of node sets
	protected $boundingBox = array();	// bounding box of nodes
	
	protected $singleBox = array();		// bounding box of single PCB
	protected $description = false;		// naming of the Gerber layer set
	
	function setDescription($description)
	{
		MKS_Sanitizer::assertNonEmptyString($description, 'Gerber layer set description');
		$this->description = $description;
	}
	
	protected function parseLayer($layer, $fileName)
	{
		MKS_Sanitizer::assertNonEmptyString($this->description, 'Gerber layer set description');
		
		$GP = new MKS_GerberParser();
		$GP->openFile($fileName);
		$GP->parsePositionalData();
		$GP->closeGaps();
		// $GP->insertVCuts();
		$this->nodeSets[$layer] = $GP->calculateDisjointSets();
		$this->boundingBox[$layer] = $GP->getBoundingBox();
		$GP->visualizeSets($this->description.' '.$this->layerName[$layer].'.png');
		
		if ($layer == 1)
		{
			for ($i = 1; $i <= $this->nodeSets[$layer]; $i++)
				$this->singleBox[$i] = $GP->getBoundingBox($i);
		}
		unset($GP);
	}
	
	function parseOutline($fileName)
	{
		$this->parseLayer(0, $fileName);
	}
	
	function parseOutlineSingle($fileName)
	{
		$this->parseLayer(1, $fileName);
	}
	
	function parseVCut($fileName)
	{
		$this->parseLayer(2, $fileName);
	}
	
	protected function getNodeSets($layer)
	{
		MKS_Sanitizer::assertNonNegativeInteger($this->nodeSets[$layer], 'number of structures on Gerber layer '.$this->layerName[$layer]);
		
		if ($this->nodeSets[$layer] == 0)
			throw new Exception('ERROR - empty Gerber layer '.$this->layerName[$layer]);
		
		return $this->nodeSets[$layer];
	}
	
	function getPanelStructuresCount()
	{
		return $this->getNodeSets(0);
	}
	
	function getSingleStructuresCount()
	{
		return $this->getNodeSets(1);
	}
	
	protected function getBoundingBox($layer)
	{
		MKS_Sanitizer::assertNonEmptyArray($this->boundingBox[$layer], 'bounding box for Gerber layer '.$this->layerName[$layer]);
		MKS_Sanitizer::assertNumberBetween(4, 4, count($this->boundingBox[$layer]), 'bounding box information for Gerber layer '.$this->layerName[$layer]);
		
		return $this->boundingBox[$layer];
	}
	
	function getPanelDimensions()
	{
		return $this->getBoundingBox(0);
	}
	
	function getPanelBorder()
	{
		$panel = $this->getBoundingBox(0);
		$singles = $this->getBoundingBox(1);
		
		foreach ($panel as $i => $val1)
			$panel[$i] = pow(-1, $i)*($singles[$i] - $val1);
		
		return $panel;
	}
	
	function getSingleSize()
	{
		MKS_Sanitizer::assertNonEmptyArray($this->singleBox, 'bounding boxes for single PCBs');
		
		$sizeX = array();
		$sizeY = array();
		foreach ($this->singleBox as $setID => $box)
		{
			// is this bounding box contained in any of the others?
			foreach ($this->singleBox as $containerID => $container)
			{
				if ($container[0] < $box[0] && $container[1] > $box[1]
				 && $container[2] < $box[2] && $container[3] > $box[3])
				{
					unset($this->singleBox[$setID]);
					continue 2;
				}
			}
			
			$sizeX[] = $box[1] - $box[0];
			$sizeY[] = $box[3] - $box[2];
		}
		
		// check single dimensions consistency
		$sizeX = array_unique($sizeX);
		$sizeY = array_unique($sizeY);
		if (count($sizeX) != 1 || count($sizeY) != 1)
			throw new Exception('ERROR - inconsistent PCB dimensions! (X values: '.implode(',',$sizeX).' - Y values: '.implode(',',$sizeY).')');
		
		return array_merge($sizeX, $sizeY);
	}
}

?>