<?php

// parses Gerber files between 100 B and 10 KB.
// currently only absolute measurements in
// milimeters and 4:4 format are supported.
class MKS_GerberParser
{
	// constants
	protected $lowerLimit = 100;		// constant: 100 B min size
	protected $upperLimit = 10000;		// constant: 10 KB max size
	protected $headerPattern = array(	// constant: Gerber header
		'G04 Layer_Color=\d+\*', // comment, ignored
		'%FSLAX44Y44\*%', // coordinate format 4:4
		'%MOMM\*%', // unit mm
		'G71\*', // I SAID MM!
		'G01\*', // linear interpolation
		'G75\*'); // multi quadrant mode
	
	// variables
	protected $fileHandle = false;		// file object
	protected $nodeList = array();		// list of nodes as index => (posX, posY, setID, link1, link2, ...)
	
	// define file to read which:
	// - must exist and be readable
	// - must be within the size limits
	function openFile($string)
	{
		MKS_Sanitizer::assertReadableFileName($string, 'Gerber file');
		MKS_Sanitizer::assertNumberBetween($this->lowerLimit, $this->upperLimit, filesize($string), 'Gerber file size');
		
		$this->fileHandle = fopen($string, 'r');
		
		MKS_Sanitizer::assertOpenFileHandle($this->fileHandle, 'Gerber file');
	}
	
	// test the file to contain the following lines.
	// expects the file pointer to be at the beginning.
	// run time: O(n)
	protected function checkForHeader()
	{
		// iterate through all expected lines
		foreach ($this->headerPattern as $pattern)
		{
			$line = fgets($this->fileHandle);
			$result = preg_match('/^'.$pattern.'[\r\n]+$/', $line);
			
			MKS_Sanitizer::assertRegexMatch('/^'.$pattern.'[\r\n]+$/', $line, 'Gerber file header');
		}
	}
	
	// returns the next positional data block split
	// into an array (data, posX, posX, operation).
	// posX and / or posY may be empty strings.
	// run time: O(n^2)
	function parsePositionalData()
	{
		MKS_Sanitizer::assertOpenFileHandle($this->fileHandle, 'Gerber file');
		rewind($this->fileHandle);
		$this->checkForHeader();
		
		$previous = 0;
		
		while ($line = fgets($this->fileHandle))
		{
			// file end
			if ($line == "M02*\r\n")
				break;
			
			$result = preg_match('/^(?:X(-?\d+))?(?:Y(-?\d+))?(?:I-?\d+)?(?:J-?\d+)?(?:D(\d\d))\*[\r\n]+$/', $line, $matches);
			MKS_Sanitizer::assertNumberBetween(0, 1, $result, 'preg_match return value in '.__METHOD__);

			// no match or no position
			if (strlen(@$matches[1]) + strlen(@$matches[2]) == 0)
				continue;
			
			// X value -> mm
			try
			{
				MKS_Sanitizer::assertNonEmptyString($matches[1], 'Gerber X position');
				$posX = intval($matches[1])/10000;
			}
			catch (Exception $e) {}
			
			// Y value -> mm
			try
			{
				MKS_Sanitizer::assertNonEmptyString($matches[2], 'Gerber Y position');
				$posY = intval($matches[2])/10000;
			}
			catch (Exception $e) {}
			
			// assure Gerber file sanity
			MKS_Sanitizer::assertNumeric($posX, 'Gerber X position');
			MKS_Sanitizer::assertNumeric($posY, 'Gerber Y position');
			MKS_Sanitizer::assertNumberBetween(1, 2, $matches[3], 'Gerber operation code');
			//echo 'node at '.$posX.'/'.$posY;
			
			// node properties
			$index = count($this->nodeList);
			$drawLine = ($matches[3] == 1);
			$link = $drawLine ? $previous : 0;
			$duplicate = false;
			
			// deduplication
			foreach ($this->nodeList as $index => $node)
			{
				// found a node on this position
				if ($posX == $node[0] && $posY == $node[1])
				{
					$duplicate = true;
					//echo ' duplicate of '.$index.'<br>';
					break;
				}
				elseif ((pow($posX - $node[0], 2) + pow($posY - $node[1], 2)) < 1 && false)
				{
					$duplicate = true;
					//echo ' close to '.$index.' ('.(pow($posX - $node[0], 2) + pow($posY - $node[1], 2)).')<br>';
					break;
				}
			}
			
			// save node
			if ($duplicate == false)
			{
				$index++;
				$this->nodeList[$index] = array($posX, $posY, 0);
				//echo ' => '.$index.'<br>';
			}
			
			// link
			if ($drawLine)
			{
				array_push($this->nodeList[$index], $link);
				array_push($this->nodeList[$link], $index);
				//echo ' &nbsp; &nbsp; linking: '.$link.' <==> '.$index.'<br>';
			}
			
			$previous = $index;
		}
		
		while ($line = fgets($this->fileHandle))
			throw new Exception('ERROR - data present after Gerber EOF code!');
	}
	
	// links nodes closer than 1mm, which ensures
	// that they get recognized as a set later on.
	// run time: O(n^2)
	function closeGaps()
	{
		MKS_Sanitizer::assertNonEmptyArray($this->nodeList, 'node list');
		
		foreach ($this->nodeList as $i1 => $n1)
		{
			for ($i2 = $i1+1; isset($this->nodeList[$i2]); $i2++)
			{
				$n2 = $this->nodeList[$i2];
				$distanceSquared = pow($n1[0] - $n2[0], 2) + pow($n1[1] - $n2[1], 2);
				
				/*if ($distanceSquared < 100) // less than 10mm
					echo 'distance between node '.$i1.' and '.$i2.': '.round(pow($distanceSquared, 0.5), 3).'mm<br>';//*/
				
				// link nodes closer than 1mm
				if ($distanceSquared < 1)
				{
					array_push($this->nodeList[$i1], $i2);
					array_push($this->nodeList[$i2], $i1);
				}
			}
		}
	}
	
	// sets the setID in the node list and
	// returns the amount of disjoint sets.
	function calculateDisjointSets()
	{
		MKS_Sanitizer::assertNonEmptyArray($this->nodeList, 'node list');
		
		$setID = 0;
		foreach ($this->nodeList as $index => $node)
		{
			// skip if already defined
			if ($this->nodeList[$index][2] > 0)
				continue;
			
			// list of node ids in this set
			$nodeIDs = array($index);
			$count_current = 1;
			
			// find all connected nodes
			do
			{
				$count_previous = $count_current;
				
				// add neighbours
				foreach ($nodeIDs as $nodeID)
					$nodeIDs = array_merge($nodeIDs, array_slice($this->nodeList[$nodeID], 3));
				
				$nodeIDs = array_unique($nodeIDs);
				$count_current = count($nodeIDs);
				
			} while ($count_current > $count_previous);
			
			// assign all connected nodes to new set
			$setID++;
			foreach ($nodeIDs as $nodeID)
				$this->nodeList[$nodeID][2] = $setID;
		}
		
		return $setID;
	}
	
	// returns an array of unique nodes each defined by an
	// array nodeID => (posX, posY, setID, link1, link2, ...).
	// only the nodes of the specified set (if given) will be returned.
	// the node indices are preserved
	function getNodeList($setID=false)
	{
		MKS_Sanitizer::assertNonEmptyArray($this->nodeList, 'node list');
		
		// setID specified
		try
		{
			MKS_Sanitizer::assertNonNegativeInteger($setID, 'node set ID');
			
			$setNodes = array();
			foreach ($this->nodeList as $index => $node)
				if ($node[2] == $setID)
					$setNodes[$index] = $node;
			
			return $setNodes;
		}
		catch (Exception $e) {}
		
		// setID not specified
		return $this->nodeList;
	}
	
	// get the min and max values for X and Y
	// as an array(minX, maxX, minY, maxY)
	// of all nodes of the given set (default all).
	// max values need to be non-negative,
	// min values need to be at most max values.
	function getBoundingBox($setID=false)
	{
		// get a non-empty node set
		$nodes = $this->getNodeList($setID);
		MKS_Sanitizer::assertNonEmptyArray($nodes, 'node list');
		
		// calculate min and max values
		$boundingBox = array(PHP_INT_MAX, -PHP_INT_MAX, PHP_INT_MAX, -PHP_INT_MAX);
		foreach ($nodes as $node)
		{
			$boundingBox[0] = min($boundingBox[0], $node[0]);
			$boundingBox[1] = max($boundingBox[1], $node[0]);
			$boundingBox[2] = min($boundingBox[2], $node[1]);
			$boundingBox[3] = max($boundingBox[3], $node[1]);
		}
		
		// check results
		MKS_Sanitizer::assertNumberBetween(0, PHP_INT_MAX, $boundingBox[1], 'bounding box border right');
		MKS_Sanitizer::assertNumberBetween(0, PHP_INT_MAX, $boundingBox[3], 'bounding box border up');
		MKS_Sanitizer::assertNumberBetween(-PHP_INT_MAX, $boundingBox[1], $boundingBox[0], 'bounding box border left');
		MKS_Sanitizer::assertNumberBetween(-PHP_INT_MAX, $boundingBox[3], $boundingBox[2], 'bounding box border down');
		
		return $boundingBox;
	}
	
	// close the file again
	function __destruct()
	{
		@fclose($this->fileHandle);
	}
	
	// --------------- //
	// debug functions //
	// --------------- //
	
	/*function calculateSetProximity()
	{
		$proximity = array();
		
		for ($setID1 = 1; count($set1 = $this->getNodeList($setID1)); $setID1++)
		{
			for ($setID2 = $setID1+1; count($set2 = $this->getNodeList($setID2)); $setID2++)
			{
				$proximity[$setID1][$setID2] = PHP_INT_MAX;
				foreach ($set1 as $i1 => $n1)
				{
					foreach ($set2 as $i2 => $n2)
					{
						$distance = pow(pow($n1[0] - $n2[0], 2) + pow($n1[1] - $n2[1], 2), 0.5);
						if ($distance < 1)
							echo 'gap closing fail between nodes '.$i1.' ('.implode(',',$n1).') and '.$i2.' ('.implode(',',$n2).')<br>';
						$proximity[$setID1][$setID2] = min($distance, $proximity[$setID1][$setID2]);
					}
				}
			}
		}
		
		return $proximity;
	} // end any possible comment */
	
	function visualizeSets($filename)
	{
		MKS_Sanitizer::assertNonEmptyString($filename, 'image file name');
		
		// general image parameters
		$boundingBox = $this->getBoundingBox();
		$sizeX = $boundingBox[1] - $boundingBox[0];
		$sizeY = $boundingBox[3] - $boundingBox[2];
		$interval = $sizeX + 5;
		
		$im = @imagecreatetruecolor(4000, $sizeY+1) or die('Cannot Initialize new GD image stream');
		$bgColor = imagecolorallocate($im, 255, 255, 255);
		$frameColor = imagecolorallocate($im, 239, 239, 239);
		$lineColor = imagecolorallocate($im, 0, 0, 0);
		imagefill($im, 0, 0, $bgColor);
		
		for ($i = 1; count($nodes = $this->getNodeList($i)); $i++)
		{
			imagefilledrectangle($im, $interval*($i-1), 0, $interval*($i-1)+$sizeX, $sizeY, $frameColor);
			foreach ($nodes as $node)
			{
				while (count($node) > 3)
				{
					$link = array_pop($node);
					imageline($im, 
						($interval*($i-1))+round($node[0]-$boundingBox[0]), 
						$sizeY-round($node[1]-$boundingBox[2]), 
						($interval*($i-1))+round($this->nodeList[$link][0]-$boundingBox[0]), 
						$sizeY-round($this->nodeList[$link][1]-$boundingBox[2]), 
						$lineColor);
				}
			}
		}
		
		imagepng($im, $filename);
	} // end any possible comment */
}

?>